// Authors: Sven Förster, Duc Thanh Tran

#include "equitableColoring.hpp"
#include "boost.hpp"
#include "randomGraphFactory.hpp"
#include "status.hpp"
#include "timer.hpp"

#include <boost/dynamic_bitset.hpp>
#include <boost/graph/push_relabel_max_flow.hpp>

#include <algorithm>
#include <iomanip>
#include <iostream>
#include <iterator>
#include <numeric>
#include <tuple>
#include <vector>

namespace EqDsatur {

using Status::to_underlying;

EquitableColoring::EquitableColoring(const Parameters &pArg) : parameters_(pArg), startTimer_(new Timer(times_.timeTotal)) {
	// read in or generate graph
	if (parameters_.graphInput == 'R') {
		graph_ = RandomGraphFactory::create(parameters_.n, parameters_.p, parameters_.randomSeed);
	}
	else if (parameters_.graphInput == 'N') {
		graph_ = Input::readGraph(parameters_.graphFile);
		parameters_.n = static_cast<int>(boost::num_vertices(graph_));
	}

	initialization();

	if (bounds_.LB != bounds_.UB) {
		bounds_.startLB = bounds_.LB;
		bounds_.startUB = bounds_.UB;

		if (parameters_.eqDsaturVariant == 'N') {
			dSatur();
		}
		else if (parameters_.eqDsaturVariant == 'C') {
			initPropertyMapsMaxFlow();
			dSaturClique();
		}
	}

	startTimer_.reset(nullptr);
	printOutput();
}

Vertex EquitableColoring::passVSS() noexcept {
	const Timer timerPassVss(times_.timeVertexSelection);

	const auto maxSatDeg = findMaxSatDeg();
	const auto vertMaxSatDeg = findVertexSatDeg(maxSatDeg);

	auto vertex = vertMaxSatDeg.front();

	if (vertMaxSatDeg.size() == 1) {
		return vertex;
	}

	if (currentStatus_.nColors - maxSatDeg > parameters_.threshold) {
		vertex = *std::max_element(vertMaxSatDeg.cbegin(), vertMaxSatDeg.cend(), [this](const Vertex &lhs, const Vertex &rhs) {
			return getDegree(lhs) < getDegree(rhs);
		});
	}
	else {
		vertex = *std::max_element(vertMaxSatDeg.cbegin(), vertMaxSatDeg.cend(), [this, &maxSatDeg](const Vertex &lhs, const Vertex &rhs) {
			return pass(lhs, maxSatDeg) < pass(rhs, maxSatDeg);
		});
	}

	return vertex;
}

int EquitableColoring::pass(const Vertex vertex, const int maxSatDeg) const noexcept {
	auto sum = 0;
	for (auto color = 1; color < bounds_.UB; ++color) {
		if (isColorAvailable(vertex, color)) {
			sum += std::accumulate(boost::adjacent_vertices(vertex, graph_).first, boost::adjacent_vertices(vertex, graph_).second, 0, [maxSatDeg, color,this](int acc, const Vertex adjVertex) {
				return isColorAvailable(adjVertex, color) && saturationDegreeMap_[adjVertex] == maxSatDeg ? ++acc : acc;
			});
		}
	}
	return sum;
}

void EquitableColoring::removeVertexFromIndependentCliques(const Vertex vertex) noexcept {
	for (auto i = 0; i < static_cast<int>(indClq_.size()); ++i) {
		for (auto j = 0; j < static_cast<int>(indClq_[i].size()); ++j) {
			if (indClq_[i][j] == vertex) {

				indClq_[i].erase(indClq_[i].begin()+j);
				--cliques_.verticesInCliques;
				cliqueIndex_[vertex] = to_underlying(CliqueIndex::RestVertex);
				cliqueMap_[vertex] = to_underlying(CliqueStatus::NoClique);

				if (indClq_[i].size() == 1) {
					const auto neighbor = indClq_[i].front();
					if(needUpdateCombPruning()) {
						moveToRestVertex(neighbor);
						cliqueIndex_[neighbor] = to_underlying(CliqueIndex::RestVertex);
						updateCliqueDecompCliqueRemoval(i+1);
						freeColors_.cliqueFreeColors.erase(freeColors_.cliqueFreeColors.begin()+1+i);
						if (needUpdateExclusiveColor()) {
							freeColors_.cliqueExclusiveFreeColors.erase(freeColors_.cliqueExclusiveFreeColors.begin()+1+i);
						}
					}

					cliqueMap_[neighbor] = to_underlying(CliqueStatus::NoClique);

					indClq_.erase(indClq_.begin()+i);
					--cliques_.verticesInCliques;
					--cliques_.numberCliques;
				}

				return;
			}
		}
	}
}

void EquitableColoring::updateCliqueDecompCliqueRemoval(const int removedCliqueIndex) noexcept {
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (getCliqueIndex(*vertex) > removedCliqueIndex) {
			--cliqueIndex_[*vertex];
		}
	}
}

void EquitableColoring::colorStartClique(const Clique &clique) noexcept {
	auto color = 1;
	for (const auto vertex : clique) {
		colorVertex(vertex, color);
		++color;
	}
}

void EquitableColoring::calculateUpperBound() noexcept {
	const auto currentStatusTemp = currentStatus_;
	const auto freeColorsTemp = freeColors_;
	const auto cliqueIndexTemp = cliqueIndex_;

	const auto colorMapOriginal = colorMap_;
	const auto colorClassCardinalityOriginal = colorClassCardinality_;
	const auto saturationDegreeMapOriginal = saturationDegreeMap_;
	const auto forbiddenColorMapOriginal = forbiddenColorMap_;
	const auto rankMapOriginal = rankMap_;

	greedyColoring();
	bounds_.UB = naiveUB();

	colorMap_ = colorMapOriginal;
	colorClassCardinality_ = colorClassCardinalityOriginal;
	saturationDegreeMap_ = saturationDegreeMapOriginal;
	forbiddenColorMap_ = forbiddenColorMapOriginal;
	rankMap_ = rankMapOriginal;
	currentStatus_ = currentStatusTemp;
	freeColors_ = freeColorsTemp;
	cliqueIndex_ = cliqueIndexTemp;
}

void EquitableColoring::greedyColoring() noexcept {
	auto colorsUsed = currentStatus_.nColors;

	const auto findSmallestPossibleColor = [this](const Vertex vertex) {
		for (auto color = 1; color <= currentStatus_.nColors; ++color) {
			if (isColorAvailable(vertex, color)) {
				return color;
			}
		}
		return -1;
	};

	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (isUncolored(*vertex)) {
			const auto smallestColor = findSmallestPossibleColor(*vertex);
			if (smallestColor == -1) {
				++colorsUsed;
				colorVertex(*vertex, colorsUsed);
			}
			else {
				colorVertex(*vertex, smallestColor);
			}
		}
	}
}

int EquitableColoring::naiveUB() noexcept {
	auto couldSwapVertex = false;
	while (!isEquitable()) {
		// find smallest and largest color class sizes
		auto cMinMax = std::make_pair(INT_MAX, INT_MIN);
		for (auto i = 0; i < currentStatus_.nColors; ++i) {
			if (colorClassCardinality_[i] > cMinMax.second) {
				cMinMax.second = colorClassCardinality_[i];
			}
			if (colorClassCardinality_[i] < cMinMax.first) {
				cMinMax.first = colorClassCardinality_[i];
			}
		}

		VertexIt vertex, v_end;
		Vertex v = static_cast<Vertex>(-1);
		for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
			v = *vertex;

			if (colorClassCardinality_[colorMap_[*vertex]-1] == cMinMax.second) {
				for (auto j = 1; j <= currentStatus_.nColors; ++j) {
					if (colorClassCardinality_[j-1] == cMinMax.first) {
						couldSwapVertex = swapNodeToColor(*vertex, j);
						if (couldSwapVertex) { break; }
					}
				}

				if (couldSwapVertex) { break; }
			}
		}

		if (!couldSwapVertex) {
			++currentStatus_.nColors;
			--colorClassCardinality_[colorMap_[v]-1];
			colorMap_[v] = currentStatus_.nColors;
			++colorClassCardinality_[currentStatus_.nColors-1];
		}
	}
	return currentStatus_.nColors;
}

bool EquitableColoring::swapNodeToColor(const Vertex vertex, const int color) noexcept {
	const auto tmpColor = colorMap_[vertex];
	colorMap_[vertex] = color;

	if (!isColoring()) {
		colorMap_[vertex] = tmpColor;
		return false;
	}
	else {
		--colorClassCardinality_[tmpColor-1];
		++colorClassCardinality_[color-1];
		return true;
	}
}

bool EquitableColoring::isColoring() const noexcept {
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		OutEdgeIt oEdge, oEdgeEnd;
		for (std::tie(oEdge, oEdgeEnd) = boost::out_edges(*vertex, graph_); oEdge != oEdgeEnd; ++oEdge) {
			if (colorMap_[boost::source(*oEdge, graph_)] == colorMap_[boost::target(*oEdge, graph_)]) {
				return false;
			}
		}
	}
	return true;
}

bool EquitableColoring::isEquitable() const noexcept {
	for (auto i = 0; i < currentStatus_.nColors; ++i) {
		for (auto j = i+1; j < currentStatus_.nColors; ++j) {
			if (std::abs(colorClassCardinality_[i] - colorClassCardinality_[j]) > 1)
				return false;
		}
	}
	return true;
}

void EquitableColoring::colorVertex(const Vertex vertex, const int color) noexcept {
	if(needUpdateCombPruning()) {
		colorVertexFreeColors(vertex, color);
	}

	++currentStatus_.rank;
	--currentStatus_.uncoloredVertices;
	colorMap_[vertex] = color;
	rankMap_[vertex] = currentStatus_.rank;

	if (color > currentStatus_.nColors) {
		++currentStatus_.nColors;
		backtrack_.fRC.emplace_back(currentStatus_.rank);
	}
	addFBC(vertex, color);
	incColorClass(color);
}

void EquitableColoring::uncolorVertex(const Vertex vertex) noexcept {
	--currentStatus_.rank;
	++currentStatus_.uncoloredVertices;

	const auto color = colorMap_[vertex];

	if(needUpdateCombPruning()) {
		uncolorVertexFreeColors(vertex, color);
	}

	colorMap_[vertex] = to_underlying(VertexColor::Uncolored);
	rankMap_[vertex] = to_underlying(RankStatus::ZeroRank);

	removeFBC(vertex, color);
	decColorClass(color);
}

void EquitableColoring::addFBC(const Vertex vertex, const int color) noexcept {
	AdjacentIt adjVertex, adjVEnd;
	for (std::tie(adjVertex, adjVEnd) = boost::adjacent_vertices(vertex, graph_); adjVertex != adjVEnd; ++adjVertex) {
		++forbiddenColorMap_[*adjVertex][color-1];

		if (forbiddenColorMap_[*adjVertex][color-1] == to_underlying(ForbiddenColorStatus::NotAvailable)) {
			++saturationDegreeMap_[*adjVertex];
		}
	}
}

void EquitableColoring::removeFBC(const Vertex vertex, const int color) noexcept {
	AdjacentIt adjVertex, adjVEnd;
	for (std::tie(adjVertex, adjVEnd) = boost::adjacent_vertices(vertex, graph_); adjVertex != adjVEnd; ++adjVertex) {
		--forbiddenColorMap_[*adjVertex][color-1];

		if (forbiddenColorMap_[*adjVertex][color-1] == to_underlying(ForbiddenColorStatus::Available)) {
			--saturationDegreeMap_[*adjVertex];
		}
	}
}

void EquitableColoring::incColorClass(const int color) noexcept {
	++colorClassCardinality_[color-1];
	updateTandM(color, true);
}

void EquitableColoring::decColorClass(const int color) noexcept {
	--colorClassCardinality_[color-1];
	if (colorClassCardinality_[color-1] == to_underlying(ColorClassStatus::Empty)) {
		--currentStatus_.nColors;
		backtrack_.fRC.pop_back();
	}
	updateTandM(color, false);
}

void EquitableColoring::updateTandM(const int lastUsedColor, const bool wasIncremented) noexcept {
	if (wasIncremented) {
		if (currentStatus_.M < colorClassCardinality_[lastUsedColor-1]) {
			currentStatus_.M = colorClassCardinality_[lastUsedColor-1];
			currentStatus_.T = 1;
		}
		else if (colorClassCardinality_[lastUsedColor-1] == currentStatus_.M) {
			++currentStatus_.T;
		}
	}
	else {
		if (colorClassCardinality_[lastUsedColor-1] == currentStatus_.M-1) {
			--currentStatus_.T;

			if (currentStatus_.T == 0) {
				--currentStatus_.M;

				auto sum = 0;
				for (auto i = 0; i < currentStatus_.nColors; ++i) {
					if (colorClassCardinality_[i] == currentStatus_.M) {
						++sum;
					}
				}
				currentStatus_.T = sum;
			}
		}
	}
}

void EquitableColoring::initialization() noexcept {
	currentStatus_.uncoloredVertices = parameters_.n;
	resizeMapsAndIndexes();
	initFreeColors();

	calculateLowerBoundAndColorStartClique();
	calculateUpperBound();

	if (parameters_.eqDsaturVariant == 'C') {
		indClq_ = computeIndependentCliques();
		if (needUpdateCombPruning()) {
			updateFreeColorsNewCliqueDecomposition();
		}
	}
}

void EquitableColoring::resizeMapsAndIndexes() noexcept {
	cliqueMap_.resize(parameters_.n, to_underlying(CliqueStatus::NoClique));
	cliqueIndex_.resize(parameters_.n, to_underlying(CliqueIndex::RestVertex));
	colorMap_.resize(parameters_.n, to_underlying(VertexColor::Uncolored));
	rankMap_.resize(parameters_.n, to_underlying(RankStatus::ZeroRank));
	colorClassCardinality_.resize(parameters_.n, 0);
	saturationDegreeMap_.resize(parameters_.n, 0);

	forbiddenColorMap_.resize(parameters_.n);
	degreeMap_.resize(parameters_.n);
	for (auto i = 0; i < parameters_.n; ++i) {
		forbiddenColorMap_[i].resize(parameters_.n, to_underlying(ForbiddenColorStatus::Available));
		degreeMap_[i] = static_cast<int>(boost::degree(i, graph_));
	}
}

void EquitableColoring::initFreeColors() noexcept {
	freeColors_.exclusiveColor.resize(parameters_.n, to_underlying(ExclusiveColor::None));

	freeColors_.cliqueFreeColors.resize(1);
	freeColors_.cliqueFreeColors[0].resize(parameters_.n, parameters_.n);
	if (needUpdateExclusiveColor()) {
		freeColors_.cliqueExclusiveFreeColors.resize(1);
		freeColors_.cliqueExclusiveFreeColors[0].resize(parameters_.n, 0);
	}
}

void EquitableColoring::calculateLowerBoundAndColorStartClique() noexcept {
	const auto startingMaxClique = findMaxClique();
	bounds_.LB = static_cast<int>(startingMaxClique.size());
	putInClique(startingMaxClique);
	colorStartClique(startingMaxClique);
}

bool EquitableColoring::putInClique(const Clique &clique) noexcept {
	++cliques_.numberCliques;
	for (const auto vertex : clique) {
		cliqueMap_[vertex] = cliques_.numberCliques;
	}
	cliques_.verticesInCliques += static_cast<int>(clique.size());
	return true;
}

int EquitableColoring::findMaxSatDeg() const noexcept {
	auto maxSatDeg = 0;
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (cliqueMap_[*vertex] != to_underlying(CliqueStatus::StartingClique)) {
			if (isUncolored(*vertex) && saturationDegreeMap_[*vertex] > maxSatDeg) {
				maxSatDeg = saturationDegreeMap_[*vertex];
			}
		}
	}
	return maxSatDeg;
}

std::vector<Vertex> EquitableColoring::findVertexSatDeg(const int satDeg) const noexcept {
	std::vector<Vertex> vertSatDeg;
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (cliqueMap_[*vertex] != to_underlying(CliqueStatus::StartingClique)) {
			if (saturationDegreeMap_[*vertex] == satDeg && isUncolored(*vertex)) {
				vertSatDeg.emplace_back(*vertex);
			}
		}
	}
	return vertSatDeg;
}

bool EquitableColoring::checkForBacktracking(const Vertex vertex) noexcept {
	for (auto color = 1; color < bounds_.UB; ++color) {
		if (isColorAvailable(vertex, color)) {
			return false;
		}
	}
	backtrack_.targetRank = backtrackToRank(vertex);
	backtrack_.doingBacktrack = true;
	++counts_.backtracks;

	return true;
}

int EquitableColoring::backtrackToRank(const Vertex vertex) const noexcept {
	std::vector<int> neighboursRankMap(currentStatus_.nColors, 0);

	AdjacentIt adjVertex, adjVEnd;
	for (std::tie(adjVertex, adjVEnd) = boost::adjacent_vertices(vertex, graph_); adjVertex != adjVEnd; ++adjVertex) {
		const auto colorAdjVertex = colorMap_[*adjVertex];
		if (colorAdjVertex != to_underlying(VertexColor::Uncolored)) {
			if (rankMap_[*adjVertex] < neighboursRankMap[colorAdjVertex-1] || neighboursRankMap[colorAdjVertex-1] == 0) {
				neighboursRankMap[colorAdjVertex-1] = rankMap_[*adjVertex];
			}
		}
	}

	auto maxRank = *std::max_element(neighboursRankMap.cbegin(), neighboursRankMap.cend());
	if (maxRank <= bounds_.LB) {
		maxRank = bounds_.LB + 1;
	}
	return maxRank;
}

void EquitableColoring::newUBBacktracking() {
	if (backtrack_.fRC.back() > bounds_.LB) {
		backtrack_.doingBacktrack = true;
		++counts_.backtracks;
		backtrack_.targetRank = backtrack_.fRC.back();
	}
	else {
		times_.timeout = true;
		backtrack_.fastEnd = true;
	}
}

void EquitableColoring::printOutput() const noexcept {
	std::cout << "File: " << parameters_.graphFile << '\n'
		<< "\tVariant = " << parameters_.eqDsaturVariant << '\n'
		<< "\tRandom Seed = " << parameters_.randomSeed << '\n'
		<< "\tExecute (Theorem 2) = " << parameters_.nPruningRule << '\n'
		<< "\tExecute (Comb. Rules) = " << parameters_.nPruningCombRule << '\n'
		<< "\tExecute (Ind. Clique Decomp.) = " << parameters_.nCliqueDecompositionRule << '\n'
		<< "\tProb. Find Better Ind. Cliques = " << parameters_.pNewCl << '\n'
		<< "\tExclusive Color Pruning (1) = " << parameters_.enableCombPruning_1 << '\n'
		<< "\tColor Cardinality Pruning (2) = " << parameters_.enableCombPruning_2 << '\n'
		<< "\tExclusive Clique Color Pruning (3) = " << parameters_.enableCombPruning_3 << '\n'
		<< "\tn = " << parameters_.n << '\n'
		<< "\tedges = " << boost::num_edges(graph_) << '\n';

	if(parameters_.graphInput == 'R')
		std::cout << "\tp = " << parameters_.p << '\n';

	std::cout << "\tStarting Bounds = " << bounds_.startLB << ", " << bounds_.startUB << '\n'
		<< "\tEnding Bounds = " << bounds_.LB << ", " << bounds_.UB << '\n';

	std::cout << "\tVisited Nodes = " << counts_.visitedNodes << '\n'
		<< "\tBacktracks = " << counts_.backtracks << '\n'
		<< "\tExecuted Max Flows = " << counts_.executedMaxFlow << '\n'
		<< "\tTotal Tries Find Better Ind. Cliques = " << counts_.betterIndCliqueDecompositionTries << '\n'
		<< "\tFound Better Ind. Cliques = " << counts_.foundBetterIndCliqueDecomposition << '\n';

	std::cout << "\tTotal Time = " << std::fixed << std::setprecision(2) << times_.timeTotal << '\n'
		<< "\tTimer Vertex Selection = " << times_.timeVertexSelection << '\n'
		<< "\tTimer FF = " << times_.timeFF << '\n'
		<< "\tTimer Build FlowNetwork = " << times_.timeBuildNetwork << '\n'
		<< "\tTimer Solve Max Flows = " << times_.timeSolveMaxFlow << '\n'
		<< "\tTimer Update Ind. Cliques = " << times_.timeUpdateIndClique << '\n'
		<< "\tTimer Try Find Better Ind. Cliques = " << times_.timeFindBetterIndCliques << '\n'
		<< "\tTimer Calculate Ind. Cliques = " << times_.timeFindIndCliques << '\n'
		<< "\tTimer Exclusive Color Pruning (1) = " << times_.timePruningComb_1 << '\n'
		<< "\tTimer Color Cardinality Pruning (2) = " << times_.timePruningComb_2 << '\n'
		<< "\tTimer Exclusive Clique Color Pruning (3) = " << times_.timePruningComb_3 << '\n'
		<< "\tTimeout = " << times_.timeout << '\n';

	std::cout << "\tPruning Theorem 1 = " << counts_.pruningTheoremOne << '\n'
		<< "\tPruning Theorem 2 = " << counts_.pruningTheoremTwo << '\n'
		<< "\tPruning Exclusive Color (1) = " << counts_.pruningComb_1 << '\n'
		<< "\tPruning Color Class Cardinality (2) = " << counts_.pruningComb_2 << '\n'
		<< "\tPruning Exclusive Color Clique (3) = " << counts_.pruningComb_3 << '\n';

	std::cout << '\n';
}

// ---------------------------------------------------------------------------------
// COMB PRUNING ORIENTED -----------------------------------------------------------
// ---------------------------------------------------------------------------------

inline bool EquitableColoring::needUpdateCombPruning() const noexcept {
	return parameters_.enableCombPruning_1 > 0 || parameters_.enableCombPruning_2 > 0 || parameters_.enableCombPruning_3 > 0;
}

inline bool EquitableColoring::needUpdateExclusiveColor() const noexcept {
	return parameters_.enableCombPruning_1 > 0 || parameters_.enableCombPruning_3 > 0;
}

void EquitableColoring::setCliqueIndices(const Clique &maxClique, const int cliqueIndex) noexcept {
	for (const auto vertex : maxClique) {
		cliqueIndex_[vertex] = cliqueIndex;
	}
}

void EquitableColoring::moveToRestVertex(const Vertex vertex) noexcept {
	if (isUncolored(vertex)) {
		for (auto color = 1; color <= bounds_.UB-1; ++color) {
			if (isColorAvailable(vertex, color)) {
				--freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];
				++freeColors_.cliqueFreeColors[0][color-1];
			}
		}

		if (needUpdateExclusiveColor() && hasExclusiveColor(vertex)) {
			const auto exclusiveColor = freeColors_.exclusiveColor[vertex];
			--freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][exclusiveColor-1];
			++freeColors_.cliqueExclusiveFreeColors[0][exclusiveColor-1];
		}
	}
}

void EquitableColoring::updateFreeColorsNewCliqueDecomposition() noexcept {
	resetCountersFreeColors();
	resizeCliqueFreeColors();

	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (isUncolored(*vertex)) {
			updateVertexFreeColors(*vertex);
		}
	}
}

void EquitableColoring::resizeCliqueFreeColors() noexcept {
	// first vector is used for restVertices, i.e. vertices not in cliques
	const auto vectorCount = 1 + static_cast<int>(indClq_.size());
	freeColors_.cliqueFreeColors.resize(vectorCount);
	if (needUpdateExclusiveColor()) {
		freeColors_.cliqueExclusiveFreeColors.resize(vectorCount);
	}

	for (auto i = 0; i < vectorCount; ++i) {
		freeColors_.cliqueFreeColors[i].resize(bounds_.UB);
		if (needUpdateExclusiveColor()) {
			freeColors_.cliqueExclusiveFreeColors[i].resize(bounds_.UB);
		}
	}
}

void EquitableColoring::resetCountersFreeColors() noexcept {
	freeColors_.cliqueFreeColors.clear();
	if (needUpdateExclusiveColor()) {
		freeColors_.cliqueExclusiveFreeColors.clear();
		std::fill(freeColors_.exclusiveColor.begin(), freeColors_.exclusiveColor.end(), to_underlying(ExclusiveColor::None));
	}
}

void EquitableColoring::updateVertexFreeColors(const Vertex vertex) noexcept {
	auto freeColors = 0;
	auto lastColor = -1;
	for (auto color = 1; color <= bounds_.UB-1; ++color) {
		if (isColorAvailable(vertex, color)) {
			++freeColors;
			++freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];
			lastColor = color;
		}
	}

	if (needUpdateExclusiveColor() && freeColors == 1) {
		++freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][lastColor-1];
		freeColors_.exclusiveColor[vertex] = lastColor;
	}
}

void EquitableColoring::colorVertexFreeColors(const Vertex vertex, const int color) noexcept {
	disableFreeColors(vertex);
	if (needUpdateExclusiveColor() && hasExclusiveColor(vertex)) {
		--freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][color-1];
		resetExclusiveColorToNone(vertex);
	}

	updateNeighborsAfterColoring(vertex, color);
}

void EquitableColoring::disableFreeColors(const Vertex vertex) noexcept {
	for (auto color = 1; color <= bounds_.UB-1; ++color) {
		if (isColorAvailable(vertex, color)) {
			--freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];
		}
	}
}

void EquitableColoring::updateNeighborsAfterColoring(const Vertex vertex, const int color) noexcept {
	AdjacentIt adjVertex, adjV_end;
	for (std::tie(adjVertex, adjV_end) = boost::adjacent_vertices(vertex, graph_); adjVertex != adjV_end; ++adjVertex) {
		if (isUncolored(*adjVertex) && isColorAvailable(*adjVertex, color)) {
			updateNeighborVertexColoring(*adjVertex, color);
		}
	}
}

void EquitableColoring::updateNeighborVertexColoring(const Vertex vertex, const int color) noexcept {
	--freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];

	if (needUpdateExclusiveColor()) {
		if (hasExclusiveColor(vertex)) {
			--freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][color-1];
			resetExclusiveColorToNone(vertex);
		}
		else {
			checkForExclusiveColorColoring(vertex, color);
		}
	}
}

void EquitableColoring::checkForExclusiveColorColoring(const Vertex vertex, const int bannedColor) noexcept {
	auto freeColors = 0;
	auto lastColor = -1;
	for (auto color = 1; color <= bounds_.UB-1; ++color) {
		if (isColorAvailable(vertex, color) && color != bannedColor) {
			if (freeColors > 1) { return; }
			++freeColors;
			lastColor = color;
		}
	}

	if (needUpdateExclusiveColor() && freeColors == 1) {
		++freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][lastColor-1];
		freeColors_.exclusiveColor[vertex] = lastColor;
	}
}

void EquitableColoring::uncolorVertexFreeColors(const Vertex vertex, const int color) noexcept {
	enableFreeColors(vertex);
	if (needUpdateExclusiveColor()) {
		checkForExclusiveColorUncoloring(vertex, color);
	}
	updateNeighborsAfterUncoloring(vertex, color);
}

void EquitableColoring::enableFreeColors(const Vertex vertex) noexcept {
	for (auto color = 1; color <= bounds_.UB-1; ++color) {
		if (isColorAvailable(vertex, color)) {
			++freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];
		}
	}
}

void EquitableColoring::updateNeighborsAfterUncoloring(const Vertex vertex, const int color) noexcept {;
	AdjacentIt adjVertex, a_end;
	for (std::tie(adjVertex, a_end) = boost::adjacent_vertices(vertex, graph_); adjVertex != a_end; ++adjVertex) {
		if (isUncolored(*adjVertex) && willBeFreeColor(*adjVertex, color)) {
			updateNeighborVertexUncoloring(*adjVertex, color);
		}
	}
}

void EquitableColoring::updateNeighborVertexUncoloring(const Vertex vertex, const int color) noexcept {
	++freeColors_.cliqueFreeColors[getCliqueIndex(vertex)][color-1];

	if (needUpdateExclusiveColor()) {
		if (hasExclusiveColor(vertex)) {
			--freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][getExclusiveColor(vertex)];
			resetExclusiveColorToNone(vertex);
		}
		else {
			checkForExclusiveColorUncoloring(vertex, color);
		}
	}
}

void EquitableColoring::checkForExclusiveColorUncoloring(const Vertex vertex, const int newColor) noexcept {
	for (auto color = 1; color <= bounds_.UB-1; ++color) {
		if (color != newColor && isColorAvailable(vertex, color)) {
			return;
		}
	}

	++freeColors_.cliqueExclusiveFreeColors[getCliqueIndex(vertex)][newColor-1];
	freeColors_.exclusiveColor[vertex] = newColor;
}


// ---------------------------------------------------------------------------------
// CLIQUE ORIENTED -----------------------------------------------------------------
// ---------------------------------------------------------------------------------

IndCliqueDecomposition EquitableColoring::computeIndependentCliques() noexcept {
	const Timer timerComputeIndCliques(times_.timeFindIndCliques);
	IndCliqueDecomposition independentCliques;

	auto cliqueIndex = 1;
	auto toVisit = currentStatus_.uncoloredVertices;
	while (toVisit > 0) {
		const auto maxClique = findMaxClique();

		if (maxClique.empty()) {
			break;
		}

		if (static_cast<int>(maxClique.size()) < cliques_.verticesInSmallestClique) {
			cliques_.verticesInSmallestClique = static_cast<int>(maxClique.size());
		}

		independentCliques.emplace_back(maxClique);

		setCliqueIndices(maxClique, cliqueIndex);
		++cliqueIndex;
		putInClique(maxClique);

		toVisit -= static_cast<int>(maxClique.size());

		AdjacentIt adjVertex, adjVEnd;
		for (const auto vertex : maxClique) {
			for (std::tie(adjVertex, adjVEnd) = boost::adjacent_vertices(vertex, graph_); adjVertex != adjVEnd; ++adjVertex) {
				if (isUncoloredNotInClique(*adjVertex)) {
					cliqueMap_[*adjVertex] = to_underlying(CliqueStatus::Ignore);
					--toVisit;
				}
			}
		}
	}

	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (cliqueMap_[*vertex] == to_underlying(CliqueStatus::Ignore)) {
			cliqueMap_[*vertex] = to_underlying(CliqueStatus::NoClique);
		}
	}

	cliques_.numberVertIndCliques = cliques_.verticesInCliques - bounds_.LB;
	return independentCliques;
}

Clique EquitableColoring::findMaxClique() const noexcept {
	Clique maxClique;
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (getDegree(*vertex) >= static_cast<int>(maxClique.size()) && isUncoloredNotInClique(*vertex)) {
			const auto tmpClique = findMaxCliqueWithVertex(*vertex, static_cast<int>(maxClique.size()));

			if (tmpClique.size() >= 2 && tmpClique.size() > maxClique.size()) {
				maxClique = std::move(tmpClique);
			}
			else if (tmpClique.size() == maxClique.size()) {
				// Sewell's Condition
				if (calculateDegreeSum(tmpClique) > calculateDegreeSum(maxClique)) {
					maxClique = std::move(tmpClique);
				}
			}
		}
	}
	return maxClique;
}

int EquitableColoring::calculateDegreeSum(const Clique &clique) const noexcept {
	return std::accumulate(clique.cbegin(), clique.cend(), 0, [this](int accumulator, const Vertex vertex) {
		return accumulator + getDegree(vertex);
	});
}

Clique EquitableColoring::findMaxCliqueWithVertex(const Vertex vertex, const int currentMaxClqSize) const noexcept {
	Clique clique{ vertex };
	clique.reserve(currentMaxClqSize);

	boost::dynamic_bitset<> candidateNeighbors(parameters_.n);

	// set uncolored neighbors of 'vertex' as possible clique-candidates
	AdjacentIt adjVertex, a_end;
	for (std::tie(adjVertex, a_end) = boost::adjacent_vertices(vertex, graph_); adjVertex != a_end; ++adjVertex) {
		if (isUncoloredNotInClique(*adjVertex) && getDegree(*adjVertex) >= currentMaxClqSize) {
			candidateNeighbors.set(*adjVertex, true);
		}
	}

	const auto searchForHighestDegCandidate = [this](const boost::dynamic_bitset<> &candidates) {
		auto vertex = candidates.find_first();
		for (auto i = candidates.find_first(); i < candidates.size(); i = candidates.find_next(i)) {
			if (getDegree(vertex) <= getDegree(i)) {
				vertex = i;
			}
		}
		return vertex;
	};

	while (candidateNeighbors.any()) {
		const auto highestDegNeighbor = searchForHighestDegCandidate(candidateNeighbors);
		clique.emplace_back(highestDegNeighbor);

		boost::dynamic_bitset<> newCandidates(parameters_.n);
		AdjacentIt adjVertex2, a_end2;
		for (std::tie(adjVertex2, a_end2) = boost::adjacent_vertices(highestDegNeighbor, graph_); adjVertex2 != a_end2; ++adjVertex2) {
			newCandidates.set(*adjVertex2, true);
		}

		candidateNeighbors &= newCandidates;
	}
	return clique;
}


void EquitableColoring::initPropertyMapsMaxFlow() noexcept {
	propertyMapsMaxFlow_.capacityMap = get(boost::edge_capacity, maxFlow_.flowGraph);
	propertyMapsMaxFlow_.revEdgeMap = get(boost::edge_reverse, maxFlow_.flowGraph);
	propertyMapsMaxFlow_.residualCapMap = get(boost::edge_residual_capacity, maxFlow_.flowGraph);
	propertyMapsMaxFlow_.refVertMap = get(ref_vertex_t(), maxFlow_.flowGraph);
}

bool EquitableColoring::dSatur() noexcept {
	if (checkOvertime()) {
		times_.timeout = true;
		return true;
	}

	if (currentStatus_.uncoloredVertices == 0) {
		if (currentStatus_.nColors < bounds_.UB) {
			if (isEquitableColoring()) {
				bounds_.UB = currentStatus_.nColors;
				newUBBacktracking();
			}
		}
		return true;
	}
	const auto vertex = passVSS();

	if (isPrunableTheoremOne()) {
		++counts_.pruningTheoremOne;
		checkForBacktracking(vertex);
		return false;
	}

	for (auto color = 1; color <= std::min(bounds_.UB-1, currentStatus_.nColors+1); ++color) {
		if (isColorAvailable(vertex, color)) {
			++counts_.visitedNodes;
			colorVertex(vertex, color);
			dSatur();

			if (times_.timeout)
				return true;

			if (backtrack_.doingBacktrack) {
				if (backtrack_.targetRank == currentStatus_.rank) {
					backtrack_.doingBacktrack = false;
				}
				else if (backtrack_.targetRank < currentStatus_.rank) {
					uncolorVertex(vertex);
					return true;
				}
			}
			uncolorVertex(vertex);
		}
	}

	checkForBacktracking(vertex);
	return false;
}

bool EquitableColoring::dSaturClique() noexcept {
	if (checkOvertime()) {
		times_.timeout = true;
		return true;
	}

	if (currentStatus_.uncoloredVertices == 0) {
		if (currentStatus_.nColors < bounds_.UB) {
			if (isEquitableColoring()) {
				bounds_.UB = currentStatus_.nColors;
				newUBBacktracking();
			}
		}
		return true;
	}
	const auto vertex = passVSS();

	if (isPrunable(vertex)) {
		return false;
	}

	for (auto color = 1; color <= std::min(currentStatus_.nColors+1, bounds_.UB-1); ++color) {
		if (isColorAvailable(vertex, color)) {
			++counts_.visitedNodes;

			colorVertex(vertex, color);
			{
				const Timer timerUpdateIndClq(times_.timeUpdateIndClique);
				updateIndepCliques(vertex);
			}

			dSaturClique();

			if (times_.timeout) {
				return true;
			}

			if (backtrack_.doingBacktrack) {
				if (backtrack_.targetRank == currentStatus_.rank) {
					backtrack_.doingBacktrack = false;
					if (foundBetterIndCliqueDecomposition()) {
						cliques_.numberVertIndCliquesColored = 0L;
					}

				}
				else if (backtrack_.targetRank < currentStatus_.rank) {
					uncolorVertex(vertex);
					return true;
				}
			}
			uncolorVertex(vertex);
		}
	}
	checkForBacktracking(vertex);
	return false;
}

// ---------------------------------------------------------------------------------
// PRUNING -------------------------------------------------------------------------
// ---------------------------------------------------------------------------------

bool EquitableColoring::isPrunable(const Vertex vertex) noexcept {

	if (isPrunableTheoremOne()) {
		// std::cout << "MD pruned\n";
		++counts_.pruningTheoremOne;
		checkForBacktracking(vertex);
		return true;
	}

	const auto executeCombPruning = counts_.visitedNodes % parameters_.nPruningCombRule == 0;

	if (parameters_.enableCombPruning_3 > 0 && executeCombPruning) {
		if (isPrunableCombPruning_3()) {
			++counts_.pruningTheoremTwo;
			checkForBacktracking(vertex);
			return true;
		}
	}


	auto coloringCanBeExtended = false;
	auto lastIndex = std::max(bounds_.LB, currentStatus_.nColors);
	while (currentStatus_.M <= std::ceil(parameters_.n*1.0 / lastIndex) && lastIndex < bounds_.UB) {
		++lastIndex;
	}

	const auto executeMaxFlow = counts_.visitedNodes % parameters_.nPruningRule == 0;

	for (auto i = std::max(bounds_.LB, currentStatus_.nColors); i < lastIndex; ++i) {

		// Combinatorial Pruning Rule (1)
		if (parameters_.enableCombPruning_1 > 0 && executeCombPruning) {
			const Timer timer(times_.timePruningComb_1);
			if (isPrunableCombPruning_1(i)) {
				if (i == lastIndex - 1) {
					++counts_.pruningTheoremTwo;
					checkForBacktracking(vertex);
					return true;
				}
				continue;
			}
		}

		// Combinatorial Pruning Rule (2)
		if (parameters_.enableCombPruning_2 > 0 && executeCombPruning) {
			const Timer timer(times_.timePruningComb_2);
			if (isPrunableCombPruning_2(i)) {
				if (i == lastIndex - 1) {
					++counts_.pruningTheoremTwo;
					checkForBacktracking(vertex);
					return true;
				}
				continue;
			}
		}

		// Max Flow Pruning
		// correct already initialized pruning parameter
		if (!parameters_.doMaxFlowPruning) {
			return false;
		}

		if (parameters_.doMaxFlowPruning && executeMaxFlow) {
			const Timer timerPrunableMaxFlow(times_.timeFF);
			initBackupGraphs(i);
			if (!isPrunableMaxFlow(i)) {
				coloringCanBeExtended = true;
				break;
			}
		}
	}

	if (!executeMaxFlow) {
		// max flow pruning was skipped due to i-th search node.
		// If combinatorial pruning rules are enabled, we could not prune.
		return false;
	}

	if (!coloringCanBeExtended) {
		++counts_.pruningTheoremTwo;
		checkForBacktracking(vertex);
		return true;
	}

	return false;
}

// ---------------------------------------------------------------------------------
// PRUNING: MENDEZ-DIAZ ET. AL  ----------------------------------------------------
// ---------------------------------------------------------------------------------

inline bool EquitableColoring::isPrunableTheoremOne() const noexcept {
	return parameters_.n < (currentStatus_.M - 1) * std::max(currentStatus_.nColors, bounds_.LB) + currentStatus_.T;
}

// ---------------------------------------------------------------------------------
// PRUNING: COMBINATORIAL RULES ----------------------------------------------------
// ---------------------------------------------------------------------------------

bool EquitableColoring::isPrunableCombPruning_1(const int colorAmount) noexcept {
	for (auto color = 1; color <= colorAmount; ++color) {
		const auto verticesWithExclColors = computeExclusiveColors(color, colorAmount);
		const auto maxCapacities = computeMaxCapacities(color, colorAmount);

		// 1 color
		if (std::get<0>(verticesWithExclColors) > std::get<0>(maxCapacities)) {
			return true;
		}

		// colorAmount-1 colors, i.e. color is banned
		if (std::get<1>(verticesWithExclColors) > std::get<1>(maxCapacities)) {
			return true;
		}

		// colorAmount colors, i.e. all colors
		if (std::get<2>(verticesWithExclColors) > std::get<2>(maxCapacities)) {
			return true;
		}
	}
	return false;
}

bool EquitableColoring::isPrunableCombPruning_2(const int colorAmount) noexcept {
	for (auto color = 1; color <= colorAmount; ++color) {
		const auto cliqueVerticesCounts = computeCliqueVerticesWithFreeColors(color, colorAmount);
		const auto restVerticesCounts = computeRestVerticesWithFreeColors(color, colorAmount);
		const auto minCapacities = computeMinCapacities(color, colorAmount);

		// 1 color
		if (std::get<0>(restVerticesCounts) + std::get<0>(cliqueVerticesCounts) < std::get<0>(minCapacities)) {
			return true;
		}

		// colorAmount-1 colors, i.e. color is banned
		if (std::get<1>(restVerticesCounts) + std::get<1>(cliqueVerticesCounts) < std::get<1>(minCapacities)) {
			return true;
		}

		// colorAmount colors, i.e. all colors
		if (std::get<2>(restVerticesCounts) + std::get<2>(cliqueVerticesCounts) < std::get<2>(minCapacities)) {
			return true;
		}
	}
	return false;
}

bool EquitableColoring::isPrunableCombPruning_3() noexcept {
	const Timer startPruningColorClq(times_.timePruningComb_3);

	for (auto color = 1; color < bounds_.UB; ++color) {
		for (auto i = 1; i < static_cast<int>(freeColors_.cliqueExclusiveFreeColors.size()); ++i) {
			// 1 color is selected
			if (freeColors_.cliqueExclusiveFreeColors[i][color-1] > 1) {
				return true;
			}

			// 1 color is banned
			auto vertices = 0;
			for (auto j = 0; j < static_cast<int>(freeColors_.cliqueExclusiveFreeColors[i].size()); ++j) {
				if (j != color-1) {
					vertices += freeColors_.cliqueExclusiveFreeColors[i][j];
				}
			}
			if (vertices > bounds_.UB-2) {
				return true;
			}

			// all colors
			const auto amountVertices = std::accumulate(freeColors_.cliqueExclusiveFreeColors[i].cbegin(), freeColors_.cliqueExclusiveFreeColors[i].cend(), 0, [color,this](int accumulator, int freeVertices) {
				return accumulator + freeVertices;
			});
			if (amountVertices > bounds_.UB-1) {
				return true;
			}
		}
	}
	return false;
}

std::tuple<int, int, int> EquitableColoring::computeExclusiveColors(const int selectedAndBannedColor, const int colorAmount) const noexcept {
	auto sum1 = 0; // 1 color
	auto sum2 = 0; // colorAmount-1 colors
	auto sum3 = 0; // colorAmount colors

	for (auto i = 0; i < static_cast<int>(freeColors_.cliqueExclusiveFreeColors.size()); ++i) {
		for (auto color = 1; color <= colorAmount; ++color) {
			if (color == selectedAndBannedColor) {
				sum1 += freeColors_.cliqueExclusiveFreeColors[i][color-1];
			}
			if (color != selectedAndBannedColor) {
				sum2 += freeColors_.cliqueExclusiveFreeColors[i][color-1];
			}
			sum3 += freeColors_.cliqueExclusiveFreeColors[i][color-1];

		}
	}
	return std::make_tuple(sum1, sum2, sum3);
}

std::tuple<int, int, int> EquitableColoring::computeMaxCapacities(const int selectedAndBannedColor, const int colorAmount) const noexcept {
	const auto capacity1 = static_cast<int>(std::ceil(parameters_.n*1.0 / colorAmount)) - getColorClassSize(selectedAndBannedColor); // 1 color
	auto capacity2 = 0; // colorAmount-1 colors
	auto capacity3 = 0; // colorAmount colors
	for (auto color = 1; color <= colorAmount; ++color) {
		const auto cap = static_cast<int>(std::ceil(parameters_.n*1.0 / colorAmount)) - getColorClassSize(color);
		if (color != selectedAndBannedColor) {
			capacity2 += cap;
		}
		capacity3 += cap;
	}
	return std::make_tuple(capacity1, capacity2, capacity3);
}

std::tuple<int, int, int> EquitableColoring::computeCliqueVerticesWithFreeColors(const int selectedAndBannedColor, const int colorAmount) const noexcept {
	auto sum1 = 0; // 1 color
	auto sum2 = 0; // colorAmount-1 colors
	auto sum3 = 0; // colorAmount colors
	for (auto i = 1; i < static_cast<int>(freeColors_.cliqueFreeColors.size()); ++i) {
		auto count1 = 0;
		auto count2 = 0;
		auto count3 = 0;
		for (auto color = 1; color <= colorAmount; ++color) {
			// only consider colors which can still be added
			if (getColorClassSize(color) != static_cast<int>(std::ceil(parameters_.n*1.0 / colorAmount))) {
				if (color == selectedAndBannedColor) {
					count1 += std::min(1, freeColors_.cliqueFreeColors[i][color-1]);
				}
				if (color != selectedAndBannedColor) {
					count2 += std::min(1, freeColors_.cliqueFreeColors[i][color-1]);
				}
				count3 += std::min(1, freeColors_.cliqueFreeColors[i][color-1]);
			}
		}
		sum1 += std::min(static_cast<int>(indClq_[i-1].size()), count1);
		sum2 += std::min(static_cast<int>(indClq_[i-1].size()), count2);
		sum3 += std::min(static_cast<int>(indClq_[i-1].size()), count3);
	}

	return std::make_tuple(sum1, sum2, sum3);
}

std::tuple<int, int, int> EquitableColoring::computeRestVerticesWithFreeColors(const int selectedAndBannedColor, const int colorAmount) const noexcept {
	const auto sum1 = getColorClassSize(selectedAndBannedColor) != static_cast<int>(std::ceil(parameters_.n*1.0 / colorAmount)) ? freeColors_.cliqueFreeColors[0][selectedAndBannedColor-1] : 0; // 1 color
	auto sum2 = 0; // colorAmount-1 colors
	auto sum3 = 0; // colorAmount colors

	for (auto color = 1; color <= colorAmount; ++color) {
		// only consider colors which can still be added
		if (getColorClassSize(color) != static_cast<int>(std::ceil(parameters_.n*1.0 / colorAmount))) {
			if (color != selectedAndBannedColor) {
				sum2 += freeColors_.cliqueFreeColors[0][color-1];
			}
			sum3 += freeColors_.cliqueFreeColors[0][color-1];
		}
	}
	return std::make_tuple(sum1, sum2, sum3);
}

std::tuple<int, int, int> EquitableColoring::computeMinCapacities(const int selectedAndBannedColor, const int colorAmount) const noexcept {
	const auto capacity1 = static_cast<int>(std::floor(parameters_.n*1.0 / colorAmount)) - getColorClassSize(selectedAndBannedColor); // 1 color
	auto capacity2 = 0; // colorAmount-1 colors
	auto capacity3 = 0; // colorAmount colors
	for (auto color = 1; color <= colorAmount; ++color) {
		const auto cap = static_cast<int>(std::floor(parameters_.n*1.0 / colorAmount)) - getColorClassSize(color);
		if (color != selectedAndBannedColor) {
			capacity2 += cap;
		}
		capacity3 += cap;
	}
	return std::make_tuple(capacity1, capacity2, capacity3);
}

// ---------------------------------------------------------------------------------
// PRUNING: MAX FLOW ---------------------------------------------------------------
// ---------------------------------------------------------------------------------

bool EquitableColoring::isPrunableMaxFlow(const int colors) noexcept {
	if (solveMaxFlow(maxFlow_.source2, maxFlow_.target2) != maxFlow_.sumLB) {
		return true;
	}

	initNetwork2(colors);

	return solveMaxFlow(maxFlow_.source1, maxFlow_.target1) != maxFlow_.uncoloredVertices;
}

long EquitableColoring::solveMaxFlow(const VertexFord source, const VertexFord target) noexcept {
	const Timer timerSolveMaxFlow(times_.timeSolveMaxFlow);
	++counts_.executedMaxFlow;
	const TraitsGraphFlow::vertex_descriptor s = source, t = target;
	return boost::push_relabel_max_flow(maxFlow_.flowGraph, s, t);
}

// ---------------------------------------------------------------------------------
// PRUNING: FLOW GRAPH -------------------------------------------------------------
// ---------------------------------------------------------------------------------

void EquitableColoring::initBackupGraphs(const int colors) noexcept {
	const Timer timerInitFlowGraph(times_.timeBuildNetwork);

	maxFlow_.clear();
	maxFlow_.uncoloredVertices = currentStatus_.uncoloredVertices;
	maxFlow_.source1 = boost::add_vertex(maxFlow_.flowGraph);
	maxFlow_.source2 = boost::add_vertex(maxFlow_.flowGraph);
	maxFlow_.target1 = boost::add_vertex(maxFlow_.flowGraph);
	maxFlow_.target2 = boost::add_vertex(maxFlow_.flowGraph);

	auto counter = 0;
	const auto sizeIndClq = indClq_.size();
	maxFlow_.cliqueVertices.resize(sizeIndClq);

	//add nodes in cliques for phase one
	for (unsigned int i = 0; i < sizeIndClq; ++i) {
		for (const auto vertex : indClq_[i]) {
			if (isUncolored(vertex)) {
				const auto flowVertex = boost::add_vertex(maxFlow_.flowGraph);
				maxFlow_.cliqueVertices[i].emplace_back(flowVertex);
				propertyMapsMaxFlow_.refVertMap[flowVertex] = static_cast<int>(vertex);
				++counter;
			}
			else {
				std::cout << "ERROR WITH INDCLQ: Vertex " << vertex << " is colored" << '\n';
			}
		}
	}

	//add remaining nodes (not in cliques)
	VertexIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
		if (isUncoloredNotInClique(*vertex)) {
			const auto flowVertex = add_vertex(maxFlow_.flowGraph);
			maxFlow_.restVertices.emplace_back(flowVertex);
			propertyMapsMaxFlow_.refVertMap[flowVertex] = static_cast<int>(*vertex);
			++counter;
		}
	}

	if (counter != maxFlow_.uncoloredVertices) {
		std::cout << "error uncolored vertices != counter\n";
		std::cout << counter << ", " << maxFlow_.uncoloredVertices << '\n';
	}

	// add nodes for colors for cliques for phase one
	maxFlow_.cliqueVerticeColors.resize(indClq_.size());
	for (unsigned int i = 0; i < sizeIndClq; ++i) {
		if (maxFlow_.cliqueVertices[i].empty()) {
			std::cout << "Clq Groesse war NULL\n";
			continue;
		}
		for (auto j = 0; j < colors; ++j) {
			maxFlow_.cliqueVerticeColors[i].emplace_back(add_vertex(maxFlow_.flowGraph));
		}
	}

	// add color nodes
	for (auto i = 0; i < colors; ++i) {
		maxFlow_.colorVertices.emplace_back(add_vertex(maxFlow_.flowGraph));
	}

	initNetwork(colors);
}

void EquitableColoring::initNetwork(const int colors) noexcept {
	const auto sizeIndClq = indClq_.size();

	// source <-> clique nodes
	for (unsigned int i = 0; i < sizeIndClq; ++i) {
		for (unsigned int j = 0; j < maxFlow_.cliqueVertices[i].size(); ++j) {
			addEdgesIntoFlowNetwork(maxFlow_.source1, maxFlow_.cliqueVertices[i][j], 1L, 0L);

			// clique nodes <-> clique colors
			for (auto k = 0; k < colors; ++k) {
				const auto flowVertex = maxFlow_.cliqueVertices[i][j];
				const auto refVertex = propertyMapsMaxFlow_.refVertMap[flowVertex];
				if (forbiddenColorMap_[refVertex][k] == to_underlying(ForbiddenColorStatus::Available)) {
					addEdgesIntoFlowNetwork(flowVertex, maxFlow_.cliqueVerticeColors[i][k], 1L, 0L);
				}
			}
		}
	}

	// source <-> remaining vert
	for (const auto vertex : maxFlow_.restVertices) {
		addEdgesIntoFlowNetwork(maxFlow_.source1, vertex, 1L, 0L);
	}

	// clique colors <-> colors
	for (unsigned int i = 0; i < sizeIndClq; ++i) {
		if (maxFlow_.cliqueVertices[i].empty()) {
			std::cout << "A1: Clique leer\n";
			continue;
		}
		for (int k = 0; k < colors; ++k) {
			addEdgesIntoFlowNetwork(maxFlow_.cliqueVerticeColors[i][k], maxFlow_.colorVertices[k], 1L, 0L);
		}
	}

	// remaining vert <-> colors
	for (const auto vertex : maxFlow_.restVertices) {
		for (int k = 0; k < colors; ++k) {
			const auto refVertex = propertyMapsMaxFlow_.refVertMap[vertex];
			if (forbiddenColorMap_[refVertex][k] == to_underlying(ForbiddenColorStatus::Available)) {
				addEdgesIntoFlowNetwork(vertex, maxFlow_.colorVertices[k], 1L, 0L);
			}
		}
	}

	// colors <-> target
	for (auto k = 0; k < colors; ++k) {
		const auto rU = std::ceil(parameters_.n*1. / colors) - colorClassCardinality_[k];
		auto rL = std::floor(parameters_.n*1. / colors) - colorClassCardinality_[k];

		if (rL < 0) {
			rL = 0;
		}
		maxFlow_.sumLB += static_cast<int>(rL);

		addEdgesIntoFlowNetwork(maxFlow_.colorVertices[k], maxFlow_.target1, static_cast<long>(rU-rL), 0L);
		addEdgesIntoFlowNetwork(maxFlow_.colorVertices[k], maxFlow_.target2, static_cast<long>(rL), 0L);
	}

	addEdgesIntoFlowNetwork(maxFlow_.target1, maxFlow_.source1, maxFlow_.uncoloredVertices, 0L);
	addEdgesIntoFlowNetwork(maxFlow_.source2, maxFlow_.target1, maxFlow_.sumLB, 0L);

}


void EquitableColoring::initNetwork2(const int colors) noexcept {
	const Timer timerInitNetwork(times_.timeBuildNetwork);
	modifyEdgesFlowNetwork(maxFlow_.target1, maxFlow_.source1, 0L, 0L);
	modifyEdgesFlowNetwork(maxFlow_.source2, maxFlow_.target1, 0L, 0L);

	for (auto k = 0; k < colors; ++k) {
		const auto rU = std::ceil(parameters_.n*1. / colors) - colorClassCardinality_[k];
		modifyEdgesFlowNetwork(maxFlow_.colorVertices[k], maxFlow_.target1, static_cast<long>(rU), 0L);
		modifyEdgesFlowNetwork(maxFlow_.colorVertices[k], maxFlow_.target2, 0L, 0L);
	}
}

void EquitableColoring::printResidualCap() const noexcept {
	VertexFordIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(maxFlow_.flowGraph); vertex != v_end; ++vertex) {
		// std::cout << *vertex << ": ";
		EdgesOutFordIt edge, e_end;
		for (std::tie(edge, e_end) = boost::out_edges(*vertex, maxFlow_.flowGraph); edge != e_end; ++edge) {
			// std::cout << propertyMapsMaxFlow_.residualCapMap[*edge] << ", ";
		}
		// std::cout << '\n';
	}
}

void EquitableColoring::resetCapacities() const noexcept {
	VertexFordIt vertex, v_end;
	for (std::tie(vertex, v_end) = boost::vertices(maxFlow_.flowGraph); vertex != v_end; ++vertex) {
		EdgesOutFordIt edge, e_end;
		for (std::tie(edge, e_end) = boost::out_edges(*vertex, maxFlow_.flowGraph); edge != e_end; ++edge) {
			propertyMapsMaxFlow_.residualCapMap[*edge] = 0L;
		}
	}
}

void EquitableColoring::addEdgesIntoFlowNetwork(const VertexFord source, const VertexFord target, long capacityFwd, long capacityBwd) noexcept {
	const auto edgeFwd = boost::add_edge(source, target, maxFlow_.flowGraph).first;
	const auto edgeBwd = boost::add_edge(target, source, maxFlow_.flowGraph).first;

	propertyMapsMaxFlow_.capacityMap[edgeFwd] = capacityFwd;
	propertyMapsMaxFlow_.capacityMap[edgeBwd] = capacityBwd;

	propertyMapsMaxFlow_.residualCapMap[edgeFwd] = 0L;
	propertyMapsMaxFlow_.residualCapMap[edgeBwd] = 0L;

	propertyMapsMaxFlow_.revEdgeMap[edgeFwd] = edgeBwd;
	propertyMapsMaxFlow_.revEdgeMap[edgeBwd] = edgeFwd;
}

void EquitableColoring::modifyEdgesFlowNetwork(const VertexFord source, const VertexFord target, long capacityFwd, long capacityBwd) noexcept {
	const auto edgeFwd = boost::edge(source, target, maxFlow_.flowGraph).first;
	const auto edgeBwd = boost::edge(target, source, maxFlow_.flowGraph).first;

	propertyMapsMaxFlow_.capacityMap[edgeFwd] = capacityFwd;
	propertyMapsMaxFlow_.capacityMap[edgeBwd] = capacityBwd;

	propertyMapsMaxFlow_.residualCapMap[edgeFwd] = 0L;
	propertyMapsMaxFlow_.residualCapMap[edgeBwd] = 0L;
}

// ---------------------------------------------------------------------------------
// Independent Clique Decomposition Related ----------------------------------------
// ---------------------------------------------------------------------------------

void EquitableColoring::updateIndepCliques(const Vertex vertex) noexcept {
	if (cliqueMap_[vertex] > to_underlying(CliqueStatus::StartingClique)) {
		++cliques_.numberVertIndCliquesColored;

		if (cliques_.numberVertIndCliquesColored / (cliques_.numberVertIndCliques*1.0) > parameters_.pNewCl && foundBetterIndCliqueDecomposition()) {
			cliques_.numberVertIndCliquesColored = 0;
		}
		else {
			removeVertexFromIndependentCliques(vertex);
		}
	}
}

bool EquitableColoring::foundBetterIndCliqueDecomposition() noexcept {
	if (counts_.visitedNodes % parameters_.nCliqueDecompositionRule == 0) {
		const Timer timerTryFindBetterIndClique(times_.timeFindBetterIndCliques);
		++counts_.betterIndCliqueDecompositionTries;

		const auto oldCliqueMap = cliqueMap_;
		const auto oldCliques = cliques_;
		const auto oldIndClq = indClq_;
		const auto oldCliqueIndex = cliqueIndex_;
		const auto oldFreeColors = freeColors_;

		// reset clique vertices to no-clique vertices in order to find new clique decomposition
		VertexIt vertex, v_end;
		for (std::tie(vertex, v_end) = boost::vertices(graph_); vertex != v_end; ++vertex) {
			if (cliqueMap_[*vertex] > to_underlying(CliqueStatus::StartingClique)) {
				cliqueMap_[*vertex] = to_underlying(CliqueStatus::NoClique);
			}
		}

		// we still have our initial clique which we do not reset
		setClique(bounds_.LB, 1);
		resetCliqueIndices();

		indClq_ = computeIndependentCliques();

		if (cliques_.verticesInCliques < oldCliques.verticesInCliques || (cliques_.verticesInCliques == oldCliques.verticesInCliques &&
																		  cliques_.verticesInSmallestClique < oldCliques.verticesInSmallestClique)) {
			// We could not find a better independent clique decomposition. Use the old one instead.
			cliqueMap_ = std::move(oldCliqueMap);
			cliques_ = std::move(oldCliques);
			indClq_ = std::move(oldIndClq);
			cliqueIndex_ = std::move(oldCliqueIndex);
			freeColors_ = std::move(oldFreeColors);
		}
		else {
			// We have found a better independent clique decomposition, therefore we eventually need to update the free-colors.
			++counts_.foundBetterIndCliqueDecomposition;
			if (needUpdateCombPruning()) {
				updateFreeColorsNewCliqueDecomposition();
			}
			return true;
		}
	}
	return false;
}

// NOTE: not used at the moment - for future implementations
// int EquitableColoring::countFreeColors(const IndCliqueDecomposition &cliqueDecomposition) const noexcept {
// 	auto freeColorCount = 0;
// 	for (const auto clique : cliqueDecomposition) {
// 		for (const auto vertex : clique) {
// 			for (auto color = 1; color <= std::min(currentStatus_.nColors+1, bounds_.UB-1); ++color) {
// 				if (isColorAvailable(vertex, color)) { ++freeColorCount; }
// 			}
// 		}
// 	}
// 	return freeColorCount;
// }

// ---------------------------------------------------------------------------------
// Utility Methods  ----------------------------------------------------------------
// ---------------------------------------------------------------------------------

inline bool EquitableColoring::checkOvertime() const noexcept {
	return startTimer_.get()->elapsed() > static_cast<double>(parameters_.timeLimit);
}

void EquitableColoring::setClique(const int verticesInCliques, const int numberCliques) noexcept {
	cliques_.verticesInCliques = verticesInCliques;
	cliques_.numberCliques = numberCliques;
	cliques_.verticesInSmallestClique = INT_MAX;
}

} // namespace EqDsatur
