// Authors: Sven Förster, Duc Thanh Tran

#ifndef STRUCTS_HPP
#define STRUCTS_HPP

#include "boost.hpp"
#include "timer.hpp"

#include <string>
#include <vector>

namespace EqDsatur {
	
struct Parameters {
	int nPruningRule{ 0 }; //!< do pruning check (max. flow problem) after i-th node
	int nPruningCombRule{ 0 }; //!< do pruning check (comb. rules) after i-th node
	int nCliqueDecompositionRule{ 0 }; //!< do clique decomposition after i-th node
	char eqDsaturVariant{ }; //!< 'N' runs Mendez-Diaz' DSATUR-algorithm, 'C' runs our algorithm
	int timeLimit{ 0 }; //!< timelimit for our algorithm
	int threshold{ 0 }; //!< threshold value according to Mendez-Diaz et. al paper, used in passVSS
	int n{ 0 }; //!< number of nodes in graph
	double p{ 0 }; //!< density which is used for random graph generation
	char graphInput{ }; //!< 'R' denotes random graph input, 'N' denotes specific graph as ressource from a file
	std::string	graphFile{ }; //!< denotes path to graph ressource file
	int randomSeed{ 0 }; //!< used for random graph generation
	int enableCombPruning_1{ 0 };
	int enableCombPruning_2{ 0 };
	int enableCombPruning_3{ 0 };
	bool doMaxFlowPruning { false };

	/// denotes percentage threshold for colored nodes in clique-decomposition. If this threshold is reached a
	/// new clique decomposition is computed. This parameter was not used in experiments; furthermore, we could
	/// tune our algorithm with this parameter.
	double pNewCl{ 0.0 };
};

struct PropertyMapsMaxFlow {
	CapacityMap capacityMap;			//!< property map to acess the capacity of an edge
	ReverseEdgeMap revEdgeMap;			//!< property map to acess the reverse edge of an edge
	ResidualCapacityMap residualCapMap;	//!< property map to acess the capacity of an edge in the corresponing residual network
	RefVertexMap refVertMap;			//!< property map to acess the corresponding vertex in the actual graph (if existant for this vertex in the backup graph)
};

/// Main structure which saves information about maximum flow graph
struct MaxFlow {
	VertexFord source1;
	VertexFord source2;
	VertexFord target1;
	VertexFord target2;
	GraphFlow flowGraph;
	std::vector<std::vector<VertexFord> > cliqueVertices;
	std::vector<VertexFord>	restVertices;
	std::vector<std::vector<VertexFord> > cliqueVerticeColors;
	std::vector<VertexFord> colorVertices;
	int uncoloredVertices;
	int	sumLB;

	void clear() {
		flowGraph.clear();
		cliqueVertices.clear();
		restVertices.clear();
		cliqueVerticeColors.clear();
		colorVertices.clear();
		sumLB = 0;
	}
};

struct CurrentStatus {
	int nColors{ 0 }; //!< current number of colors
	int rank{ 0 }; //!< current number of nodes which are colored
	int uncoloredVertices{ 0 };

	//! notations come from paper
	int M{ 0 };	//!< number of nodes in largest color class
	int T{ 0 };	//!< number of color classes which contain exactly M nodes, i.e. number of largest color classes
};

struct Bounds {
	int UB{ 0 }; //!< current upper bound
	int LB{ 0 }; //!< current lower bound
	int startUB{ 0 };
	int startLB{ 0 };
};

struct Counts {
	unsigned long long int visitedNodes{ 0 };
	unsigned long long int backtracks{ 0 };
	unsigned long long int executedMaxFlow{ 0 };
	unsigned long long int betterIndCliqueDecompositionTries{ 0 };
	unsigned long long int foundBetterIndCliqueDecomposition{ 0 };
	unsigned long long int pruningTheoremOne{ 0 };
	unsigned long long int pruningTheoremTwo{ 0 };
	unsigned long long int pruningComb_1{ 0 };
	unsigned long long int pruningComb_2{ 0 };
	unsigned long long int pruningComb_3{ 0 };
};

/// Saves information about current independent clique decomposition
struct Cliques {
	int verticesInCliques{ 0 };
	int numberCliques{ 0 };
	int numberVertIndCliquesColored{ 0 };
	int numberVertIndCliques{ 0 };
	int verticesInSmallestClique{ INT_MAX };
};

/// Saves backtracking information
struct Backtracking {
	Vertex toNode{ 0 }; //!< target node of backtracking process
	int targetRank{ 0 }; //!< rank of the node where the algorithm heads to - note: not used, can be deleted
	bool fastEnd{ false };
	bool doingBacktrack{ false };
	std::vector<int> fRC;
};

struct Times {
	double timeTotal{ 0.0 };
	double timeVertexSelection{ 0.0 };
	double timeFF{ 0.0 };
	double timeBuildNetwork{ 0.0 };
	double timeSolveMaxFlow{ 0.0 };
	double timeUpdateIndClique{ 0.0 };
	double timeFindBetterIndCliques{ 0.0 };
	double timeFindIndCliques{ 0.0 };
	double timePruningComb_1{ 0.0 };
	double timePruningComb_2{ 0.0 };
	double timePruningComb_3{ 0.0 };
	bool timeout{ false };
};

/// Saves free color and exclusive color information. Used for combinatorial pruning rules.
struct FreeColors {
	std::vector<std::vector<int> > cliqueFreeColors;
	std::vector<std::vector<int> > cliqueExclusiveFreeColors;
	std::vector<int> exclusiveColor;
};

} // namespace EqDsatur

#endif
